/****************************************************/
/*                  Serveur de diffusion            */
/****************************************************/

#include "message.h"

int sock_rm, sock_dc, acc_rm; /* descripteurs de sockets */
int reuse=1;  /* port reutilisable */
t_message message;
int nbclients=0;  /* indique le nombre de clients connectes */
char *thosts[32]; /* tableau contenant les noms des clients */
char nom_dclient[20]; /* nom du client diffuseur */
char mess_diff[20];   /* message  diffuser */

struct sockaddr_in serveur_rm; /* socket de reception de messages */
struct sockaddr_in serveur_dc; /* socket de diffusion de message */
struct hostent *hp;


extern void emettre_message(int client);
int recevoir_message(int client);
void diffuser(t_message msg);
void fermer_connexions();
void attente_diffuseur();
void traitant_FIN();

/**************************************************************************/
/*                                  diffuser()                            */
/* fonction permettant de diffuser un message a tous les client connectes */
/**************************************************************************/

void diffuser(t_message msg)
{
   int i;
   for (i=0 ; i<32 ; i++)
     {
       if (thosts[i]!=NULL) {
         /*strcpy(message.mess,msg);*/
         emettre_message(i);
       }
     }
}

/**************************************************************************/
/*                         fermer_connexions()                             */
/* fonction permettant de fermer toutes les connexions avec les clients   */
/**************************************************************************/

void fermer_connexions()
{
   int i, c;
   for (i=0 ; i<32 ; i++)
     {
       if (thosts[i]!=NULL)
         {
           close(i);
           free(thosts[i]);
           thosts[i]=NULL;
         }
     }
   nbclients=0;
}

/**************************************************************************/
/*                           attente_diffuseur()                          */
/*     fonction mettant le serveur en attente d'un client diffuseur       */
/**************************************************************************/

void attente_diffuseur()
{
   int lg;

   printf("\nAttente de connexion d'un client diffuseur ...\n");

   lg=sizeof(serveur_rm);
   acc_rm=accept(sock_rm, (struct sockaddr *)&serveur_rm, (int *)&lg);
   printf("Valeur de acc_rm : %d\n", acc_rm);
   if (acc_rm==-1)
     {
       perror("Erreur accept : ");
       exit(1);
     }
}

/***********************************************************/
/*                      traitant_FIN()                     */
/*   fonction permettant d'arreter le serveur proprement   */
/***********************************************************/

void traitant_FIN(int n)
{
   message.type = D_SERVEUR;
   diffuser(message);
   close(sock_dc);
   close(sock_rm);
   close(acc_rm);
   printf("\n\nBye Bye ...\n\n");
   exit(1);
}

/**********************************************************/
/*                  recevoir_message()                    */
/* fonction permettant de recevoir un message d'un client */
/**********************************************************/

int recevoir_message(int client)
{
   int test;

   test=read(client, &message, sizeof(message));
   if (test == -1)
     {
       perror("Erreur read : ");
       exit(1);
     }
   return (int)test;
}

/***********************************************************/
/*                   programme principal                   */
/***********************************************************/

void main()
{
   /* declaration des variables & structures */
   int lg, c, test, acc_dc, retval, i;
   struct timeval tv;
   char nom[20];
   fd_set fdmask;
   struct sigaction action;

   /* installation d'un handler pour le signal SIGINT */
   signal(SIGINT, traitant_FIN);

   gethostname(nom, sizeof(nom));
   printf("L'identite du serveur est : %s\n",nom);
   hp=gethostbyname(nom);

   /* creer la socket d'ecoute pour communiquer avec les clients */
   sock_dc=socket(AF_INET, SOCK_STREAM, 0); /* socket de demande de connexion des clients */
   if (sock_dc==-1)
     {
       perror("Erreur socket_dc : ");
       exit(1);
     }
   else
     printf("Valeur de sock_dc : %d\n", sock_dc);

   /* creer la socket d'ecoute pour communiquer avec le client_diffuseur */
   sock_rm=socket(AF_INET, SOCK_STREAM, 0); /* socket de reception de messages */
   if (sock_rm==-1)
     {
       perror("Erreur socket_rm : ");
       exit(1);
     }
   else
     printf("Valeur de sock_rm : %d\n", sock_rm);

   /* initialisation de la structure serveur_rm cote diffuseur */
   bzero(&serveur_rm, sizeof(serveur_rm));
   serveur_rm.sin_family=AF_INET;
   serveur_rm.sin_port=htons(PORT_RM);
   bcopy(hp->h_addr, (char *)&serveur_rm.sin_addr, hp->h_length);

   /* initialisation de la structure serveur_dc cote client */
   bzero(&serveur_dc, sizeof(serveur_dc));
   serveur_dc.sin_family=AF_INET;
   serveur_dc.sin_port=htons(PORT_DC);

   /* attachement de sock_rm */
   setsockopt(sock_rm, SOL_SOCKET, SO_REUSEADDR, (int *)&reuse, sizeof(reuse));
   if (bind(sock_rm, (struct sockaddr *)&serveur_rm, sizeof(serveur_rm))==-1)
     {
       perror("Erreur bind_rm : ");
       exit(1);
     }

   /* attachement de sock_dc */
   setsockopt(sock_dc, SOL_SOCKET, SO_REUSEADDR, (int *)&reuse, sizeof(reuse));
   if (bind(sock_dc, (struct sockaddr *)&serveur_dc, sizeof(serveur_dc))==-1)
     {
       perror("Erreur bind_dc : ");
       exit(1);
     }

   /* creation file d'attente de la demande de connexion du diffuseur */
   if (listen(sock_rm, 1)==-1)
     {
       perror("Erreur listen_rm : ");
       exit(1);
     }

   /* creation file d'attente des demandes de connexion des clients */
   if (listen(sock_dc, 5)==-1)
     {
       perror("Erreur listen_dc : ");
       exit(1);
     }

    /* attente de la connexion d'un diffuseur */
   attente_diffuseur();
   printf("Demande de connexion d'un diffuseur !\n");

   /* boucle infinie du serveur */
   while (1)
   {
     FD_ZERO(&fdmask); /* effacer tous les descripteurs associes a fdmask */
     FD_SET(sock_dc, &fdmask); /* ajout du descripteurs sock_dc */
     FD_SET(acc_rm, &fdmask); /* ajout du descripteur sock_rm */

   /* ajout des descripteurs de tous les clients connectes */
     for (i=0 ; i<32 ; i++)
     {
       /* printf("Je rentre dans for ...\n"); */
       if (thosts[i]!=NULL) {
         printf("Valeur de thosts[%i] : %s\n",i,thosts[i]);
         FD_SET(i, &fdmask);
         }
     }

   /* ecoute de tous les descripteurs de sockets */
   /* pendant 15 secondes maxi */

     tv.tv_sec = 15;
     tv.tv_usec = 0;

     retval=select(32, &fdmask, NULL, NULL, &tv);
     if ( retval < 0)
     {
       printf("Erreur select number %i: \n", errno);
       perror("Erreur select ");
       exit(1);
     }
     printf("Donnes disponibles maintenant, %d descripteurs ...\n",retval);

   /* si reception du client diffuseur */
     if (FD_ISSET(acc_rm, &fdmask))
     {
       printf("1er appel recevoir_message ...\n\n");
       test=recevoir_message(acc_rm);
       printf("Nombre de caractres du message de client_diff %d\n",test);

   /* test si reception info du client diffuseur */
       if (message.type==INFO_DCLIENT)
       {
         printf("\nLe client diffuseur %s vient de se connecter ...\n\n", message.mess);
         strcpy(nom_dclient, message.mess);
         printf("2e appel recevoir_message ...\n\n");
         test=recevoir_message(acc_rm);
         strcpy(mess_diff, message.mess);
       }

   /* test si message a diffuser a tous les clients */
       if (message.type==M_DCLIENT)
       {
         printf("Le message  diffuser est : %s\n", message.mess);
         diffuser(message);
         /*test=recevoir_message(acc_rm);*/
         strcpy(mess_diff, message.mess);
       }

   /* test si demande de deconnexion du client diffuseur */
       if (message.type==D_DCLIENT || test==0)
       {
         printf("Le client diffuseur %s vient de se deconnecter ...\n", nom_dclient);
         message.type=D_DCLIENT;
         fermer_connexions();
         close(acc_rm);
         attente_diffuseur();
       }
     }


   /* si demande de connexion d'un client sur la socket d'ecoute */
     if (FD_ISSET(sock_dc, &fdmask))
     {

   /* creer un descripteur de socket pour les clients qui demandent une connexion puis etablissement de la connexion */

       printf("\nAttente de connexion d'un client rcepteur ...\n");

       lg=sizeof(serveur_dc);
       acc_dc=accept(sock_dc, (struct sockaddr *)&serveur_dc, (int *)&lg);
       printf("Valeur de acc_dc : %d\n", acc_dc);

       if (acc_dc==-1)
       {
         perror ("Erreur accept sur sock_dc : ");
         exit(1);
       }

       printf("3e appel recevoir_message ...\n\n");
       test=recevoir_message(acc_dc);

       if (message.type==C_CLIENT)
       {
         printf("Le message  recevoir est : %s\n", mess_diff);
         /*diffuser(message);*/
         thosts[acc_dc]=(char *)malloc(strlen(message.mess)*sizeof(char));
         strcpy(thosts[acc_dc], message.mess);
         nbclients++;
         printf("Le client %s est connecte ...\n", thosts[acc_dc]);
         printf("Il y a %d clients connectes ...\n\n", nbclients);
         message.type=INFO_SERVEUR;
         strcpy(message.mess, nom);
         emettre_message(acc_dc);
         message.type=INFO_DCLIENT;
         strcpy(message.mess, nom_dclient);
         emettre_message(acc_dc);
         message.type=M_DCLIENT;
         strcpy(message.mess, mess_diff);
         emettre_message(acc_dc);
       }
     }

   /* si reception d'une dconnexion d'un client */
   for (i=0 ; i<32 ; i++)
     if (thosts[i]!=NULL)
       if (FD_ISSET(i, &fdmask))
         {
           printf("4e appel recevoir_message ...\n\n");
           test=recevoir_message(i);
           if (message.type==D_CLIENT || test==0)
             {
               printf("\nLe client %s est deconnecte.\n\n", thosts[i]);
               thosts[i]=NULL;
               close(i);
               nbclients--;
               printf("\nIl y a %d clients connectes.\n\n", nbclients);
               free(thosts[i]);
             }
         }
   }
}

